from typing import Optional, Dict

from pykotor.resource.formats.ncs.ncs_data import NCS, NCSInstruction, NCSByteCode, NCSInstructionQualifier, \
    NCSInstructionType, NCSInstructionTypeValue
from pykotor.resource.type import ResourceReader, SOURCE_TYPES, ResourceWriter, TARGET_TYPES


class NCSBinaryReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._ncs: Optional[NCS] = None
        self._instructions: Dict[int, NCSInstruction] = {}
        self._jumps: Dict[NCSInstruction, int] = {}

    def load(
            self,
            auto_close: bool = True
    ) -> NCS:
        self._ncs = NCS()

        file_type = self._reader.read_string(4)
        file_version = self._reader.read_string(4)

        if file_type != "NCS ":
            raise TypeError("The file type that was loaded is invalid.")

        if file_version != "V1.0":
            raise TypeError("The NCS version that was loaded is not supported.")

        self._instructions = {}  # offset -> instruction

        self._reader.seek(13)
        while self._reader.remaining() > 0:
            offset = self._reader.position()
            self._instructions[offset] = self._read_instruction()

        for instruction, jumpToOffset in self._jumps.items():
            instruction.jump = self._instructions[jumpToOffset]

        self._ncs.instructions = list(self._instructions.values())
        
        if auto_close:
            self._reader.close()

        return self._ncs

    def _read_instruction(self) -> NCSInstruction:
        byte_code = NCSByteCode(self._reader.read_uint8())
        qualifier = self._reader.read_uint8()
        type_value = NCSInstructionTypeValue(byte_code, qualifier)

        instruction = NCSInstruction()
        instruction.ins_type = NCSInstructionType(type_value)

        if instruction.ins_type in [NCSInstructionType.CPDOWNSP, NCSInstructionType.CPTOPSP,
                                    NCSInstructionType.CPDOWNBP, NCSInstructionType.CPTOPBP]:
            instruction.args.extend([self._reader.read_int32(big=True),
                                     self._reader.read_uint16(big=True)])

        elif instruction.ins_type in [NCSInstructionType.CONSTI]:
            instruction.args.extend([self._reader.read_uint32(big=True)])

        elif instruction.ins_type in [NCSInstructionType.CONSTF]:
            instruction.args.extend([self._reader.read_single(big=True)])

        elif instruction.ins_type in [NCSInstructionType.CONSTS]:
            length = self._reader.read_uint16(big=True)
            instruction.args.extend([self._reader.read_string(length)])

        elif instruction.ins_type in [NCSInstructionType.CONSTO]:
            instruction.args.extend([self._reader.read_uint16(big=True)])

        elif instruction.ins_type in [NCSInstructionType.ACTION]:
            instruction.args.extend([self._reader.read_uint16(big=True), self._reader.read_uint8(big=True)])

        elif instruction.ins_type in [NCSInstructionType.MOVSP]:
            instruction.args.extend([self._reader.read_int32(big=True)])

        elif instruction.ins_type in [NCSInstructionType.JMP, NCSInstructionType.JSR,
                                      NCSInstructionType.JZ, NCSInstructionType.JNZ]:
            jumpOffset = self._reader.read_int32(big=True) + self._reader.position() - 6
            self._jumps[instruction] = jumpOffset

        elif instruction.ins_type in [NCSInstructionType.DESTRUCT]:
            instruction.args.extend([self._reader.read_uint16(big=True), self._reader.read_int16(big=True), self._reader.read_uint16(big=True)])

        elif instruction.ins_type in [NCSInstructionType.DECISP, NCSInstructionType.INCISP,
                                      NCSInstructionType.DECIBP, NCSInstructionType.INCIBP]:
            instruction.args.extend([self._reader.read_uint32(big=True)])

        elif instruction.ins_type in [NCSInstructionType.STORE_STATE]:
            instruction.args.extend([self._reader.read_uint32(big=True), self._reader.read_uint32(big=True)])

        elif instruction.ins_type in [NCSInstructionType.EQUALTT, NCSInstructionType.NEQUALTT]:
            instruction.args.extend([self._reader.read_uint16])

        elif instruction.ins_type in [NCSInstructionType.NOP, NCSInstructionType.RETN, NCSInstructionType.SAVEBP,
                                      NCSInstructionType.RESTOREBP, NCSInstructionType.ADDII, NCSInstructionType.RSADDI,
                                      NCSInstructionType.RSADDO, NCSInstructionType.NEGI, NCSInstructionType.RSADDS,
                                      NCSInstructionType.EQUALII]:
            ...

        else:
            raise Exception(f"Tried to read unsupported instruction '{instruction.ins_type.name}' to NCS")

        return instruction


class NCSBinaryWriter(ResourceWriter):
    def __init__(
            self,
            ncs: NCS,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._ncs: NCS = ncs
        self._offsets: Dict[NCSInstruction, int] = {}
        self._sizes: Dict[NCSInstruction, int] = {}

    def write(
            self,
            auto_close: bool = True
    ) -> None:
        offset = 13
        for instruction in self._ncs.instructions:
            self._sizes[instruction] = self.determine_size(instruction)
            self._offsets[instruction] = offset
            offset += self._sizes[instruction]

        self._writer.write_string("NCS ")
        self._writer.write_string("V1.0")
        self._writer.write_uint8(0x42)
        self._writer.write_uint32(offset, big=True)

        for instruction in self._ncs.instructions:
            self._write_instruction(instruction)

        if auto_close:
            self._writer.close()

    def determine_size(self, instruction: NCSInstruction) -> int:
        size = 2

        if instruction.ins_type in [NCSInstructionType.CPDOWNSP, NCSInstructionType.CPTOPSP,
                                    NCSInstructionType.CPDOWNBP, NCSInstructionType.CPTOPBP]:
            size = 8

        elif instruction.ins_type in [NCSInstructionType.CONSTI]:
            size += 4

        elif instruction.ins_type in [NCSInstructionType.CONSTF]:
            size += 4

        elif instruction.ins_type in [NCSInstructionType.CONSTS]:
            size += 2 + len(instruction.args[0])

        elif instruction.ins_type in [NCSInstructionType.CONSTO]:
            size = 6

        elif instruction.ins_type in [NCSInstructionType.ACTION]:
            size = 5

        elif instruction.ins_type in [NCSInstructionType.MOVSP]:
            size += 4

        elif instruction.ins_type in [NCSInstructionType.JMP, NCSInstructionType.JSR,
                                      NCSInstructionType.JZ, NCSInstructionType.JNZ]:
            size += 4

        elif instruction.ins_type in [NCSInstructionType.DESTRUCT]:
            size += 6

        elif instruction.ins_type in [NCSInstructionType.DECISP, NCSInstructionType.INCISP,
                                      NCSInstructionType.DECIBP, NCSInstructionType.INCIBP]:
            size += 4

        elif instruction.ins_type in [NCSInstructionType.STORE_STATE]:
            size += 8

        elif instruction.ins_type in [NCSInstructionType.EQUALTT, NCSInstructionType.NEQUALTT]:
            size += 2

        elif instruction.ins_type in [NCSInstructionType.NOP]:
            ...

        else:
            ...

        return size

    def _write_instruction(self, instruction: NCSInstruction) -> None:
        self._writer.write_uint8(int(instruction.ins_type.value.byte_code))
        self._writer.write_uint8(int(instruction.ins_type.value.qualifier))

        if instruction.ins_type in [NCSInstructionType.CPDOWNSP, NCSInstructionType.CPTOPSP,
                                    NCSInstructionType.CPDOWNBP, NCSInstructionType.CPTOPBP]:
            self._writer.write_int32(instruction.args[0], big=True)
            self._writer.write_uint16(4, big=True)  # TODO: 12 for float support

        elif instruction.ins_type in [NCSInstructionType.CONSTI]:
            self._writer.write_int32(instruction.args[0], big=True)

        elif instruction.ins_type in [NCSInstructionType.CONSTF]:
            self._writer.write_single(instruction.args[0], big=True)

        elif instruction.ins_type in [NCSInstructionType.CONSTS]:
            self._writer.write_string(instruction.args[0], big=True, prefix_length=2)

        elif instruction.ins_type in [NCSInstructionType.CONSTO]:
            self._writer.write_uint32(instruction.args[0], big=True)

        elif instruction.ins_type in [NCSInstructionType.ACTION]:
            self._writer.write_uint16(instruction.args[0], big=True)
            self._writer.write_uint8(instruction.args[1], big=True)

        elif instruction.ins_type in [NCSInstructionType.MOVSP]:
            self._writer.write_int32(instruction.args[0], big=True)

        elif instruction.ins_type in [NCSInstructionType.JMP, NCSInstructionType.JSR,
                                      NCSInstructionType.JZ, NCSInstructionType.JNZ]:
            relative = self._offsets[instruction.jump] - self._offsets[instruction]
            self._writer.write_int32(relative, big=True)

        elif instruction.ins_type in [NCSInstructionType.DESTRUCT]:
            self._writer.write_uint16(instruction.args[0], big=True)
            self._writer.write_int16(instruction.args[1], big=True)
            self._writer.write_uint16(instruction.args[2], big=True)

        elif instruction.ins_type in [NCSInstructionType.DECISP, NCSInstructionType.INCISP,
                                      NCSInstructionType.DECIBP, NCSInstructionType.INCIBP]:
            self._writer.write_int32(instruction.args[0], big=True)

        elif instruction.ins_type in [NCSInstructionType.STORE_STATE]:
            self._writer.write_uint32(instruction.args[0], big=True)
            self._writer.write_uint32(instruction.args[1], big=True)

        elif instruction.ins_type in [NCSInstructionType.EQUALTT, NCSInstructionType.NEQUALTT]:
            self._writer.write_uint16(instruction.args[0], big=True)

        elif instruction.ins_type in [NCSInstructionType.EQUALII, NCSInstructionType.EQUALFF,
                                      NCSInstructionType.EQUALFF, NCSInstructionType.EQUALOO,
                                      NCSInstructionType.EQUALEFFEFF, NCSInstructionType.EQUALEVTEVT,
                                      NCSInstructionType.EQUALEVTEVT, NCSInstructionType.EQUALLOCLOC,
                                      NCSInstructionType.EQUALTALTAL, NCSInstructionType.EQUALSS]:
            ...

        elif instruction.ins_type in [NCSInstructionType.NEQUALII, NCSInstructionType.NEQUALFF,
                                      NCSInstructionType.NEQUALFF, NCSInstructionType.NEQUALOO,
                                      NCSInstructionType.NEQUALEFFEFF, NCSInstructionType.NEQUALEVTEVT,
                                      NCSInstructionType.NEQUALEVTEVT, NCSInstructionType.NEQUALLOCLOC,
                                      NCSInstructionType.NEQUALTALTAL, NCSInstructionType.NEQUALSS]:
            ...

        elif instruction.ins_type in [NCSInstructionType.ADDII, NCSInstructionType.ADDFF,
                                      NCSInstructionType.ADDFI, NCSInstructionType.ADDIF,
                                      NCSInstructionType.ADDSS, NCSInstructionType.ADDVV]:
            ...

        elif instruction.ins_type in [NCSInstructionType.SUBII, NCSInstructionType.SUBFF,
                                      NCSInstructionType.SUBFI, NCSInstructionType.SUBIF,
                                      NCSInstructionType.SUBVV]:
            ...

        elif instruction.ins_type in [NCSInstructionType.MULII, NCSInstructionType.MULFF,
                                      NCSInstructionType.MULFI, NCSInstructionType.MULIF,
                                      NCSInstructionType.MULFV, NCSInstructionType.MULVF]:
            ...

        elif instruction.ins_type in [NCSInstructionType.DIVII, NCSInstructionType.DIVFF,
                                      NCSInstructionType.DIVFI, NCSInstructionType.DIVIF,
                                      NCSInstructionType.DIVFV, NCSInstructionType.DIVVF]:
            ...

        elif instruction.ins_type in [NCSInstructionType.GTII, NCSInstructionType.GTFF,
                                      NCSInstructionType.GEQII, NCSInstructionType.GEQFF]:
            ...

        elif instruction.ins_type in [NCSInstructionType.LTII, NCSInstructionType.LTFF,
                                      NCSInstructionType.LEQII, NCSInstructionType.LEQFF]:
            ...

        elif instruction.ins_type in [NCSInstructionType.LOGANDII, NCSInstructionType.LOGORII]:
            ...

        elif instruction.ins_type in [NCSInstructionType.BOOLANDII]:
            ...

        elif instruction.ins_type in [NCSInstructionType.INCORII]:
            ...

        elif instruction.ins_type in [NCSInstructionType.NEGI, NCSInstructionType.NEGF]:
            ...

        elif instruction.ins_type in [NCSInstructionType.MODII]:
            ...

        elif instruction.ins_type in [NCSInstructionType.NOTI]:
            ...

        elif instruction.ins_type in [NCSInstructionType.RETN]:
            ...

        elif instruction.ins_type in [NCSInstructionType.RSADDI, NCSInstructionType.RSADDF, NCSInstructionType.RSADDO,
                                      NCSInstructionType.RSADDS, NCSInstructionType.RSADDEFF, NCSInstructionType.RSADDEVT,
                                      NCSInstructionType.RSADDLOC, NCSInstructionType.RSADDTAL, NCSInstructionType.SAVEBP]:
            ...

        elif instruction.ins_type in [NCSInstructionType.NOP]:
            ...

        else:
            raise Exception(f"Tried to write unsupported instruction ({instruction.ins_type.name}) to NCS")




from unittest import TestCase

from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.utw import construct_utw, dismantle_utw

TEST_FILE = "../../files/test.utw"


class TestUTW(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        utw = construct_utw(gff)
        self.validate_io(utw)

        gff = dismantle_utw(utw)
        utw = construct_utw(gff)
        self.validate_io(utw)

    def validate_io(self, utw):
        self.assertEqual(1, utw.appearance_id)
        self.assertEqual("", utw.linked_to)
        self.assertEqual("sw_mapnote011", utw.resref)
        self.assertEqual("MN_106PER2", utw.tag)
        self.assertEqual(76857, utw.name.stringref)
        self.assertEqual(-1, utw.description.stringref)
        self.assertTrue(utw.has_map_note)
        self.assertEqual(76858, utw.map_note.stringref)
        self.assertEqual(1, utw.map_note_enabled)
        self.assertEqual(5, utw.palette_id)
        self.assertEqual("comment", utw.comment)



"""
This module holds classes relating to read and write operations.
"""
from __future__ import annotations

import io
import struct
from abc import ABC, abstractmethod
from typing import BinaryIO, Union, Optional

from pykotor.common.geometry import Vector3, Vector4, Vector2
from pykotor.common.language import LocalizedString


def _endian_char(
        big
) -> str:
    """
    Returns the character that represents either big endian or small endian in struct unpack.

    Args:
        big: True if big endian.

    Returns:
        Character representing either big or small endian.
    """
    return '>' if big else '<'


class ArrayHead:
    def __init__(
            self,
            array_offset: int = 0,
            array_length: int = 0
    ):
        self.length: int = array_length
        self.offset: int = array_offset


class BinaryReader:
    """
    Used for easy reading of binary files.
    """

    def __init__(
            self,
            stream: BinaryIO,
            offset: int = 0,
            size: int = None
    ):
        self._stream: BinaryIO = stream
        self._offset: int = offset
        self.auto_close: bool = True
        self._stream.seek(offset)

        available = self.true_size() - offset
        self._size: int = available if size is None else size
        if available > self.true_size():
            raise IOError("Specified size is greater than the number of available bytes.")

    def __enter__(
            self
    ):
        return self

    def __exit__(
            self,
            exc_type,
            exc_val,
            exc_tb
    ):
        if self.auto_close: self.close()

    @classmethod
    def from_file(
            cls,
            path: str,
            offset: int = 0,
            size: int = None
    ) -> BinaryReader:
        """
        Returns a new BinaryReader with a stream established to the specified path.

        Args:
            path: Path of the file to open.
            offset: Number of bytes into the stream to consider as position 0.
            size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

        Returns:
            A new BinaryReader instance.
        """
        stream = open(path, 'rb')
        return BinaryReader(stream, offset, size)

    @classmethod
    def from_bytes(
            cls,
            data: bytes,
            offset: int = 0,
            size: int = None
    ) -> BinaryReader:
        """
        Returns a new BinaryReader with a stream established to the bytes stored in memory.

        Args:
            data: The bytes of data.
            offset: Number of bytes into the stream to consider as position 0.
            size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

        Returns:
            A new BinaryReader instance.
        """
        stream = io.BytesIO(data)
        return BinaryReader(stream, offset, size)

    @classmethod
    def from_auto(
            cls,
            source: Optional[str, bytes, bytearray, BinaryReader],
            offset: int = 0,
            size: int = None
    ):
        if isinstance(source, str):  # is path
            reader = BinaryReader.from_file(source, offset, size)
        elif isinstance(source, bytes) or isinstance(source, bytearray):  # is binary data
            reader = BinaryReader.from_bytes(source, offset, size)
        elif isinstance(source, BinaryReader):
            reader = BinaryReader(source._stream, source._offset, source._size)
        else:
            raise NotImplementedError("Must specify a path, bytes-like object or an existing BinaryReader instance.")

        return reader

    @staticmethod
    def load_file(
            path: str,
            offset: int = 0,
            size: int = -1
    ) -> bytes:
        """
        Returns bytes of a file at from specified path.

        Args:
            path: The path of the file.
            offset: The offset into the file.
            size: The amount of bytes to load, if size equals -1 loads the whole file.

        Returns:
            The bytes of the file.
        """
        with open(path, 'rb') as reader:
            reader.seek(offset)
            if size == -1:
                return reader.read()
            else:
                return reader.read(size)

    def offset(
            self
    ) -> int:
        return self._offset

    def set_offset(
            self,
            offset: int
    ) -> None:
        self.seek(self.position() + offset)
        self._offset = offset

    def size(
            self
    ) -> int:
        """
        Returns the total number of bytes accessible.

        Returns:
            The number of accessible bytes.
        """
        return self._size

    def true_size(
            self
    ) -> int:
        """
        Returns the total number of bytes in the stream.

        Returns:
            The total file size.
        """
        pos = self._stream.tell()
        self._stream.seek(0, 2)
        size = self._stream.tell()
        self._stream.seek(pos)
        return size

    def remaining(
            self
    ) -> int:
        """
        Returns the remaining number of bytes in the stream.

        Returns:
            The total file size.
        """
        return self.size() - self.position()

    def close(
            self
    ) -> None:
        """
        Closes the stream.
        """
        self._stream.close()

    def skip(
            self,
            length
    ) -> None:
        """
        Skips ahead in the stream the specified number of bytes.

        Args:
            length: How many bytes to skip.
        """
        self.exceed_check(length)
        self._stream.read(length)

    def position(
            self
    ) -> int:
        """
        Returns the byte offset into the stream.

        Returns:
            The byte offset.
        """
        return self._stream.tell() - self._offset

    def seek(
            self,
            position
    ) -> None:
        """
        Moves the stream pointer to the byte offset.

        Args:
            position: The byte index into stream.
        """
        self.exceed_check(position - self.position())
        self._stream.seek(position + self._offset)

    def read_all(
            self
    ) -> bytes:
        length = self.size() - self._offset
        self._stream.seek(self._offset)
        return self._stream.read(length)

    def read_uint8(
            self,
            *,
            big: bool = False
    ) -> int:
        """
        Reads an unsigned 8-bit integer from the stream.

        Args:
            big: Read int bytes as big endian.

        Returns:
            An integer from the stream.
        """
        self.exceed_check(1)
        return struct.unpack(_endian_char(big) + 'B', self._stream.read(1))[0]

    def read_int8(
            self,
            *,
            big: bool = False
    ) -> int:
        """
        Reads an signed 8-bit integer from the stream.

        Args:
            big: Read int bytes as big endian.

        Returns:
            An integer from the stream.
        """
        self.exceed_check(1)
        return struct.unpack(_endian_char(big) + 'b', self._stream.read(1))[0]

    def read_uint16(
            self,
            *,
            big: bool = False
    ) -> int:
        """
        Reads an unsigned 16-bit integer from the stream.

        Args:
            big: Read int bytes as big endian.

        Returns:
            An integer from the stream.
        """
        self.exceed_check(2)
        return struct.unpack(_endian_char(big) + 'H', self._stream.read(2))[0]

    def read_int16(
            self,
            *,
            big: bool = False
    ) -> int:
        """
        Reads an signed 16-bit integer from the stream.

        Args:
            big: Read int bytes as big endian.

        Returns:
            An integer from the stream.
        """
        self.exceed_check(2)
        return struct.unpack(_endian_char(big) + 'h', self._stream.read(2))[0]

    def read_uint32(
            self,
            *,
            max_neg1: bool = False,
            big: bool = False
    ) -> int:
        """
        Reads an unsigned 32-bit integer from the stream.

        If max_is_neg1 flag is set to true and the bytes read off the stream are equal to 0xFFFFFFFF then the method
        will return a value of -1 instead of 4294967295.

        Args:
            max_neg1: Return -1 when the value of the stream equals 0xFFFFFFFF.
            big: Read int bytes as big endian.

        Returns:
            An integer from the stream.
        """
        self.exceed_check(4)
        unpacked = struct.unpack(_endian_char(big) + "I", self._stream.read(4))[0]

        if unpacked == 4294967295 and max_neg1:
            unpacked = -1

        return unpacked

    def read_int32(
            self,
            *,
            big: bool = False
    ) -> int:
        """
        Reads an signed 32-bit integer from the stream.

        Args:
            big: Read int bytes as big endian.

        Returns:
            An integer from the stream.
        """
        self.exceed_check(4)
        return struct.unpack(_endian_char(big) + 'i', self._stream.read(4))[0]

    def read_uint64(
            self,
            *,
            big: bool = False
    ) -> int:
        """
        Reads an unsigned 64-bit integer from the stream.

        Args:
            big: Read int bytes as big endian.

        Returns:
            An integer from the stream.
        """
        self.exceed_check(8)
        return struct.unpack(_endian_char(big) + 'Q', self._stream.read(8))[0]

    def read_int64(
            self,
            *,
            big: bool = False
    ) -> int:
        """
        Reads an signed 64-bit integer from the stream.

        Args:
            big: Read int bytes as big endian.

        Returns:
            An integer from the stream.
        """
        self.exceed_check(8)
        return struct.unpack(_endian_char(big) + 'q', self._stream.read(8))[0]

    def read_single(
            self,
            *,
            big: bool = False
    ) -> int:
        """
        Reads an 32-bit floating point number from the stream.

        Args:
            big: Read float bytes as big endian.

        Returns:
            An float from the stream.
        """
        self.exceed_check(4)
        return struct.unpack(_endian_char(big) + 'f', self._stream.read(4))[0]

    def read_double(
            self,
            *,
            big: bool = False
    ) -> int:
        """
        Reads an 64-bit floating point number from the stream.

        Args:
            big: Read float bytes as big endian.

        Returns:
            An float from the stream.
        """
        self.exceed_check(8)
        return struct.unpack(_endian_char(big) + 'd', self._stream.read(8))[0]

    def read_vector2(
            self,
            *,
            big: bool = False
    ) -> Vector2:
        """
        Reads a two 32-bit floating point numbers from the stream.

        Args:
            big: Read bytes as big endian.

        Returns:
            A new Vector2 instance using floats read from the stream.
        """
        self.exceed_check(8)
        x, y = self.read_single(big=big), self.read_single(big=big)
        return Vector2(x, y)

    def read_vector3(
            self,
            *,
            big: bool = False
    ) -> Vector3:
        """
        Reads a three 32-bit floating point numbers from the stream.

        Args:
            big: Read bytes as big endian.

        Returns:
            A new Vector3 instance using floats read from the stream.
        """
        self.exceed_check(12)
        x, y, z = self.read_single(big=big), self.read_single(big=big), self.read_single(big=big)
        return Vector3(x, y, z)

    def read_vector4(
            self,
            *,
            big: bool = False
    ) -> Vector4:
        """
        Reads a four 32-bit floating point numbers from the stream.

        Args:
            big: Read bytes as big endian.

        Returns:
            A new Vector4 instance using floats read from the stream.
        """
        self.exceed_check(16)
        x, y, z, w = self.read_single(big=big), self.read_single(big=big), self.read_single(big=big), \
                     self.read_single(big=big)
        return Vector4(x, y, z, w)

    def read_bytes(
            self,
            length: int
    ) -> bytes:
        """
        Reads a specified number of bytes from the stream.

        Args:
            length: Number of bytes to read.

        Returns:
            A bytes object containing the read bytes.
        """
        self.exceed_check(length)
        return self._stream.read(length)

    def read_string(
            self,
            length: int,
            encoding: str = "windows-1252"
    ) -> str:
        """
        Reads a string from the stream with the specified length. Any null bytes and characters proceeding a null byte
        are trimmed from the final value and any unknown characters are ignored.

        Args:
            length: Amount of character to read.
            encoding: Encoding of string to read.

        Returns:
            A string read from the stream.
        """
        self.exceed_check(length)
        string = self._stream.read(length).decode(encoding, errors='ignore')
        if '\0' in string:
            string = string[:string.index('\0')].rstrip('\0')
            string = string.replace('\0', '')
        return string

    def read_terminated_string(
            self,
            terminator: str
    ) -> str:
        """
        Reads a string continuously from the stream until it hits the terminator string specified. Any unknown
        characters are ignored.

        Args:
            terminator: The terminator string.

        Returns:
            A string read from the stream.
        """
        string = ""
        char = ""
        while char != terminator:
            string += char
            self.exceed_check(1)
            char = self.read_bytes(1).decode('ascii', errors='ignore')
        return string

    def read_locstring(
            self
    ) -> LocalizedString:
        """
        Reads the localized string data structure from the stream.

        The binary data structure that is read follows the structure found in the GFF format specification.

        Returns:
            A LocalizedString read from the stream.
        """
        locstring = LocalizedString.from_invalid()
        self.skip(4)  # total number of bytes of the localized string
        locstring.stringref = self.read_uint32(max_neg1=True)
        string_count = self.read_uint32()
        for i in range(string_count):
            string_id = self.read_uint32()
            length = self.read_uint32()
            string = self.read_string(length)
            language, gender = LocalizedString.substring_pair(string_id)
            locstring.set(language, gender, string)
        return locstring

    def read_array_head(
            self
    ) -> ArrayHead:
        return ArrayHead(self.read_uint32(), self.read_uint32())

    def peek(
            self,
            length: int = 1
    ) -> bytes:
        data = self._stream.read(length)
        self._stream.seek(-length, 1)
        return data

    def exceed_check(
            self,
            num: int
    ) -> None:
        """
        Raises an error if the specified number exceeds the number of remaining bytes.

        Args:
            num: Number of bytes.

        Raises:
            IOError: If the given number sex exceeds the number of remaining bytes.
        """
        if self.position() + num > self.size():
            raise IOError("This operation would exceeed the streams boundaries.")


class BinaryWriter(ABC):
    @classmethod
    def to_file(
            cls,
            path: str
    ) -> BinaryWriter:
        """
        Returns a new BinaryWriter with a stream established to the specified path.

        Args:
            path: Path of the file to open.

        Returns:
            A new BinaryWriter instance.
        """
        stream = open(path, 'wb')
        return BinaryWriterFile(stream)

    @classmethod
    def to_bytearray(
            cls,
            data: bytearray = None
    ) -> BinaryWriter:
        """
        Returns a new BinaryWriter with a stream established to the specified bytes.

        Args:
            data: The bytes to write to.

        Returns:
            A new BinaryWriter instance.
        """
        if data is None:
            data = bytearray()
        return BinaryWriterBytearray(data)

    @classmethod
    def to_auto(
            cls,
            source: Union[str, bytearray, BinaryWriter]
    ) -> BinaryWriter:
        if isinstance(source, str):  # is path
            return BinaryWriter.to_file(source)
        elif isinstance(source, bytearray):  # is binary data
            return BinaryWriter.to_bytearray(source)
        elif isinstance(source, BinaryWriter):
            return source
        else:
            raise NotImplementedError("Must specify a path, bytes object or an existing BinaryWriter instance.")

    @staticmethod
    def dump(
            path: str,
            data: bytes
    ) -> None:
        """
        Convenience method used to writes the specified data to the specified file.

        Args:
            path: The filepath of the file.
            data: The data to write to the file.
        """
        with open(path, 'wb') as file:
            file.write(data)

    @abstractmethod
    def close(
            self
    ) -> None:
        """
        Closes the stream.
        """

    @abstractmethod
    def size(
            self
    ) -> int:
        """
        Returns the total file size.

        Returns:
            The total file size.
        """

    @abstractmethod
    def data(
            self
    ) -> bytes:
        """
        Returns the full file data.

        Returns:
            The full file data.
        """

    @abstractmethod
    def clear(
            self
    ) -> None:
        """
        Clears all the data in the file.
        """

    @abstractmethod
    def seek(
            self,
            position
    ) -> None:
        """
        Moves the stream pointer to the byte offset.

        Args:
            position: The byte index into stream.
        """

    @abstractmethod
    def end(
            self
    ) -> None:
        """
        Moves the pointer for the stream to the end.
        """

    @abstractmethod
    def position(
            self
    ) -> int:
        """
        Returns the byte offset into the stream.

        Returns:
            The byte offset.
        """

    @abstractmethod
    def write_uint8(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 8-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_int8(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 8-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_uint16(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 16-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_int16(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 16-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_uint32(
            self,
            value: int,
            *,
            max_neg1: bool = False,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 32-bit integer to the stream.

        If the max_neg1 flag is set to true and the specified value is equal to -1 then the stream will accept the value
        and write 0xFFFFFFFF to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
            max_neg1: When the value is -1 it is to be converted to the max uint32 value.
        """

    @abstractmethod
    def write_int32(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 32-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_uint64(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 64-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_int64(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 64-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_single(
            self,
            value: float,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an 32-bit floating point number to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_double(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a 64-bit floating point number to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """

    @abstractmethod
    def write_vector2(
            self,
            value: Vector2,
            *,
            big: bool = False
    ) -> None:
        """
        Writes two 32-bit floating point numbers to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """

    @abstractmethod
    def write_vector3(
            self,
            value: Vector3,
            *,
            big: bool = False
    ) -> None:
        """
        Writes three 32-bit floating point numbers to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """

    @abstractmethod
    def write_vector4(
            self,
            value: Vector4,
            *,
            big: bool = False
    ) -> None:
        """
        Writes four 32-bit floating point numbers to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """

    @abstractmethod
    def write_bytes(
            self,
            value: bytes
    ) -> None:
        """
        Writes the specified bytes to the stream.

        Args:
            value: The bytes to be written.
        """

    @abstractmethod
    def write_string(
            self,
            value: str,
            encoding: str = "windows-1252",
            *,
            big: bool = False,
            prefix_length: int = 0,
            string_length: int = -1,
            padding: str = '\0'
    ) -> None:
        """
        Writes the specified string to the stream. The string can also be prefixed by an integer specifying the
        strings length.

        Args:
            value: The string to be written.
            encoding: The string encoding.
            prefix_length: The number of bytes for the string length prefix. Valid options are 0, 1, 2 and 4.
            big: Write the prefix length integer as big endian.
            string_length: Fixes the string length to this size, truncating or padding where necessary. Ignores if -1.
            padding: What character is used as padding where applicable.
        """

    @abstractmethod
    def write_line(
            self,
            indent: int,
            *args
    ) -> None:
        """
        Writes a line with specified indentation and array of values that are separated by whitespace.

        Args:
            indent: Level of indentation.
            *args: Values to write.
        """

    @abstractmethod
    def write_locstring(
            self,
            value: LocalizedString,
            *,
            big: bool = False
    ):
        """
        Writes the specified localized string to the stream.

        The binary data structure that is read follows the structure found in the GFF format specification.

        Args:
            value: The localized string to be written.
            big: Write any integers as big endian.
        """


class BinaryWriterFile(BinaryWriter):
    def __init__(
            self,
            stream: BinaryIO,
            offset: int = 0
    ):
        self._stream: BinaryIO = stream
        self.offset: int = offset
        self.auto_close: bool = True
        self._stream.seek(offset)

    def __enter__(
            self
    ):
        return self

    def __exit__(
            self,
            exc_type,
            exc_val,
            exc_tb
    ):
        if self.auto_close: self.close()

    def close(
            self
    ) -> None:
        """
        Closes the stream.
        """
        self._stream.close()

    def size(
            self
    ) -> int:
        """
        Returns the total file size.

        Returns:
            The total file size.
        """
        pos = self._stream.tell()
        self._stream.seek(0, 2)
        size = self._stream.tell()
        self._stream.seek(pos)
        return size

    def data(
            self
    ) -> bytes:
        """
        Returns the full file data.

        Returns:
            The full file data.
        """
        pos = self._stream.tell()
        self._stream.seek(0)
        data = self._stream.read()
        self._stream.seek(pos)
        return data

    def clear(
            self
    ) -> None:
        """
        Clears all the data in the file.
        """
        self._stream.seek(0)
        self._stream.truncate()

    def seek(
            self,
            position
    ) -> None:
        """
        Moves the stream pointer to the byte offset.

        Args:
            position: The byte index into stream.
        """
        self._stream.seek(position + self.offset)

    def end(
            self
    ) -> None:
        """
        Moves the pointer for the stream to the end.
        """
        self._stream.seek(0, 2)

    def position(
            self
    ) -> int:
        """
        Returns the byte offset into the stream.

        Returns:
            The byte offset.
        """
        return self._stream.tell() - self.offset

    def write_uint8(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 8-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'B', value))

    def write_int8(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 8-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'b', value))

    def write_uint16(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 16-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'H', value))

    def write_int16(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 16-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'h', value))

    def write_uint32(
            self,
            value: int,
            *,
            max_neg1: bool = False,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 32-bit integer to the stream.

        If the max_neg1 flag is set to true and the specified value is equal to -1 then the stream will accept the value
        and write 0xFFFFFFFF to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
            max_neg1: When the value is -1 it is to be converted to the max uint32 value.
        """
        if max_neg1 and value == -1:
            value = 4294967295

        self._stream.write(struct.pack(_endian_char(big) + 'I', value))

    def write_int32(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 32-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'i', value))

    def write_uint64(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 64-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'Q', value))

    def write_int64(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 64-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'q', value))

    def write_single(
            self,
            value: float,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an 32-bit floating point number to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'f', value))

    def write_double(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a 64-bit floating point number to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'd', value))

    def write_vector2(
            self,
            value: Vector2,
            *,
            big: bool = False
    ) -> None:
        """
        Writes two 32-bit floating point numbers to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'f', value.x))
        self._stream.write(struct.pack(_endian_char(big) + 'f', value.y))

    def write_vector3(
            self,
            value: Vector3,
            *,
            big: bool = False
    ) -> None:
        """
        Writes three 32-bit floating point numbers to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'f', value.x))
        self._stream.write(struct.pack(_endian_char(big) + 'f', value.y))
        self._stream.write(struct.pack(_endian_char(big) + 'f', value.z))

    def write_vector4(
            self,
            value: Vector4,
            *,
            big: bool = False
    ) -> None:
        """
        Writes four 32-bit floating point numbers to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._stream.write(struct.pack(_endian_char(big) + 'f', value.x))
        self._stream.write(struct.pack(_endian_char(big) + 'f', value.y))
        self._stream.write(struct.pack(_endian_char(big) + 'f', value.z))
        self._stream.write(struct.pack(_endian_char(big) + 'f', value.w))

    def write_bytes(
            self,
            value: bytes
    ) -> None:
        """
        Writes the specified bytes to the stream.

        Args:
            value: The bytes to be written.
        """
        self._stream.write(value)

    def write_string(
            self,
            value: str,
            encoding: str = "windows-1252",
            *,
            big: bool = False,
            prefix_length: int = 0,
            string_length: int = -1,
            padding: str = '\0'
    ) -> None:
        """
        Writes the specified string to the stream. The string can also be prefixed by an integer specifying the
        strings length.

        Args:
            value: The string to be written.
            encoding: The encoding of the string to be written.
            prefix_length: The number of bytes for the string length prefix. Valid options are 0, 1, 2 and 4.
            big: Write the prefix length integer as big endian.
            string_length: Fixes the string length to this size, truncating or padding where necessary. Ignores if -1.
            padding: What character is used as padding where applicable.
        """
        if prefix_length == 1:
            if len(value) > 255:
                raise ValueError("The string length is too large for a prefix length of 1.")
            self.write_uint8(len(value), big=big)
        elif prefix_length == 2:
            if len(value) > 65535:
                raise ValueError("The string length is too large for a prefix length of 2.")
            self.write_uint16(len(value), big=big)
        elif prefix_length == 4:
            if len(value) > 4294967295:
                raise ValueError("The string length is too large for a prefix length of 4.")
            self.write_uint32(len(value), big=big)
        elif prefix_length != 0:
            raise ValueError("An invalid prefix length was provided.")

        if string_length != -1:
            while len(value) < string_length:
                value += padding
            value = value[:string_length]

        self._stream.write(value.encode(encoding))

    def write_line(
            self,
            indent: int,
            *args
    ) -> None:
        """
        Writes a line with specified indentation and array of values that are separated by whitespace.

        Args:
            indent: Level of indentation.
            *args: Values to write.
        """
        line = "  " * indent
        for arg in args:
            if isinstance(arg, float):
                line += str(round(arg, 7))
            else:
                line += str(arg)
            line += " "
        line += "\n"
        self._stream.write(line.encode())

    def write_locstring(
            self,
            value: LocalizedString,
            *,
            big: bool = False
    ):
        """
        Writes the specified localized string to the stream.

        The binary data structure that is read follows the structure found in the GFF format specification.

        Args:
            value: The localized string to be written.
            big: Write any integers as big endian.
        """
        bw = BinaryWriter.to_bytearray()
        bw.write_uint32(value.stringref, big=big, max_neg1=True)
        bw.write_uint32(len(value), big=big)
        for language, gender, substring in value:
            string_id = LocalizedString.substring_id(language, gender)
            bw.write_uint32(string_id, big=big)
            bw.write_string(substring, prefix_length=4)
        locstring_data = bw.data()

        self.write_uint32(len(locstring_data))
        self.write_bytes(locstring_data)


class BinaryWriterBytearray(BinaryWriter):
    def __init__(
            self,
            ba: bytearray,
            offset: int = 0
    ):
        self._ba = ba
        self._offset: int = offset
        self._position = 0

    def __enter__(
            self
    ):
        return self

    def __exit__(
            self,
            exc_type,
            exc_val,
            exc_tb
    ):
        ...

    def close(
            self
    ) -> None:
        """
        Closes the stream.
        """

    def size(
            self
    ) -> int:
        """
        Returns the total file size.

        Returns:
            The total file size.
        """
        return len(self._ba)

    def data(
            self
    ) -> bytes:
        """
        Returns the full file data.

        Returns:
            The full file data.
        """
        return bytes(self._ba)

    def clear(
            self
    ) -> None:
        """
        Clears all the data in the file.
        """
        self._ba.clear()

    def seek(
            self,
            position
    ) -> None:
        """
        Moves the stream pointer to the byte offset.

        Args:
            position: The byte index into stream.
        """
        self._position = position

    def end(
            self
    ) -> None:
        """
        Moves the pointer for the stream to the end.
        """
        self._position = len(self._ba)

    def position(
            self
    ) -> int:
        """
        Returns the byte offset into the stream.

        Returns:
            The byte offset.
        """
        return self._position - self._offset

    def write_uint8(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 8-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._ba[self._position:self._position + 1] = struct.pack(_endian_char(big) + 'B', value)
        self._position += 1

    def write_int8(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 8-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._ba[self._position:self._position + 1] = struct.pack(_endian_char(big) + 'b', value)
        self._position += 1

    def write_uint16(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 16-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._ba[self._position:self._position + 2] = struct.pack(_endian_char(big) + 'H', value)
        self._position += 2

    def write_int16(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 16-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._ba[self._position:self._position + 2] = struct.pack(_endian_char(big) + 'h', value)
        self._position += 2

    def write_uint32(
            self,
            value: int,
            *,
            max_neg1: bool = False,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 32-bit integer to the stream.

        If the max_neg1 flag is set to true and the specified value is equal to -1 then the stream will accept the value
        and write 0xFFFFFFFF to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
            max_neg1: When the value is -1 it is to be converted to the max uint32 value.
        """
        if max_neg1 and value == -1:
            value = 4294967295
        self._ba[self._position:self._position + 4] = struct.pack(_endian_char(big) + 'I', value)
        self._position += 4

    def write_int32(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 32-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._ba[self._position:self._position + 4] = struct.pack(_endian_char(big) + 'i', value)
        self._position += 4

    def write_uint64(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an unsigned 64-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._ba[self._position:self._position + 8] = struct.pack(_endian_char(big) + 'Q', value)
        self._position += 8

    def write_int64(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a signed 64-bit integer to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._ba[self._position:self._position + 8] = struct.pack(_endian_char(big) + 'q', value)
        self._position += 8

    def write_single(
            self,
            value: float,
            *,
            big: bool = False
    ) -> None:
        """
        Writes an 32-bit floating point number to the stream.

        Args:
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._ba[self._position:self._position + 4] = struct.pack(_endian_char(big) + 'f', value)
        self._position += 4

    def write_double(
            self,
            value: int,
            *,
            big: bool = False
    ) -> None:
        """
        Writes a 64-bit floating point number to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._ba[self._position:self._position + 8] = struct.pack(_endian_char(big) + 'd', value)
        self._position += 8

    def write_vector2(
            self,
            value: Vector2,
            *,
            big: bool = False
    ) -> None:
        """
        Writes two 32-bit floating point numbers to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._ba[self._position:self._position + 4] = struct.pack(_endian_char(big) + 'f', value.x)
        self._ba[self._position + 4:self._position + 8] = struct.pack(_endian_char(big) + 'f', value.y)
        self._position += 8

    def write_vector3(
            self,
            value: Vector3,
            *,
            big: bool = False
    ) -> None:
        """
        Writes three 32-bit floating point numbers to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._ba[self._position:self._position + 4] = struct.pack(_endian_char(big) + 'f', value.x)
        self._ba[self._position + 4:self._position + 8] = struct.pack(_endian_char(big) + 'f', value.y)
        self._ba[self._position + 8:self._position + 12] = struct.pack(_endian_char(big) + 'f', value.z)
        self._position += 12

    def write_vector4(
            self,
            value: Vector4,
            *,
            big: bool = False
    ) -> None:
        """
        Writes four 32-bit floating point numbers to the stream.

        Args:
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._ba[self._position:self._position + 4] = struct.pack(_endian_char(big) + 'f', value.x)
        self._ba[self._position + 4:self._position + 8] = struct.pack(_endian_char(big) + 'f', value.y)
        self._ba[self._position + 8:self._position + 12] = struct.pack(_endian_char(big) + 'f', value.z)
        self._ba[self._position + 12:self._position + 16] = struct.pack(_endian_char(big) + 'f', value.w)
        self._position += 16

    def write_bytes(
            self,
            value: bytes
    ) -> None:
        """
        Writes the specified bytes to the stream.

        Args:
            value: The bytes to be written.
        """
        self._ba[self._position:self._position + len(value)] = value
        self._position += len(value)

    def write_string(
            self,
            value: str,
            encoding: str = "windows-1252",
            *,
            big: bool = False,
            prefix_length: int = 0,
            string_length: int = -1,
            padding: str = '\0'
    ) -> None:
        """
        Writes the specified string to the stream. The string can also be prefixed by an integer specifying the
        strings length.

        Args:
            value: The string to be written.
            prefix_length: The number of bytes for the string length prefix. Valid options are 0, 1, 2 and 4.
            big: Write the prefix length integer as big endian.
            string_length: Fixes the string length to this size, truncating or padding where necessary. Ignores if -1.
            padding: What character is used as padding where applicable.
        """
        if prefix_length == 1:
            if len(value) > 255:
                raise ValueError("The string length is too large for a prefix length of 1.")
            self.write_uint8(len(value), big=big)
        elif prefix_length == 2:
            if len(value) > 65535:
                raise ValueError("The string length is too large for a prefix length of 2.")
            self.write_uint16(len(value), big=big)
        elif prefix_length == 4:
            if len(value) > 4294967295:
                raise ValueError("The string length is too large for a prefix length of 4.")
            self.write_uint32(len(value), big=big)
        elif prefix_length != 0:
            raise ValueError("An invalid prefix length was provided.")

        if string_length != -1:
            while len(value) < string_length:
                value += padding
            value = value[:string_length]

        encoded = value.encode(encoding)
        self._ba[self._position:self._position + len(encoded)] = encoded
        self._position += len(encoded)

    def write_line(
            self,
            indent: int,
            *args
    ) -> None:
        """
        Writes a line with specified indentation and array of values that are separated by whitespace.

        Args:
            indent: Level of indentation.
            *args: Values to write.
        """
        line = "  " * indent
        for arg in args:
            if isinstance(arg, float):
                line += str(round(arg, 7))
            else:
                line += str(arg)
            line += " "
        line += "\n"

        encoded = line.encode('ascii')
        self._ba[self._position:self._position + len(encoded)] = encoded
        self._position += len(encoded)

    def write_locstring(
            self,
            value: LocalizedString,
            *,
            big: bool = False
    ) -> None:
        """
        Writes the specified localized string to the stream.

        The binary data structure that is read follows the structure found in the GFF format specification.

        Args:
            value: The localized string to be written.
            big: Write any integers as big endian.
        """
        bw = BinaryWriter.to_bytearray()
        bw.write_uint32(value.stringref, big=big, max_neg1=True)
        bw.write_uint32(len(value), big=big)

        for language, gender, substring in value:
            string_id = LocalizedString.substring_id(language, gender)
            bw.write_uint32(string_id, big=big)
            bw.write_string(substring, prefix_length=4)
        locstring_data = bw.data()

        self.write_uint32(len(locstring_data))
        self.write_bytes(locstring_data)



from unittest import TestCase

from pykotor.common.geometry import Vector3
from pykotor.common.script import DataType
from pykotor.resource.formats.ncs.compiler.interpreter import Stack


class TestStack(TestCase):

    def test_peek_past_vector(self):
        stack = Stack()
        stack.add(DataType.FLOAT, 1.0)                      # -20
        stack.add(DataType.VECTOR, Vector3(2.0, 3.0, 4.0))  # -16
        stack.add(DataType.FLOAT, 5.0)                      # -4
        print(stack.peek(-20))

    def test_move_negative(self):
        stack = Stack()
        stack.add(DataType.FLOAT, 1)  # -24
        stack.add(DataType.FLOAT, 2)  # -20
        stack.add(DataType.FLOAT, 3)  # -16
        stack.add(DataType.FLOAT, 4)  # -12
        stack.add(DataType.FLOAT, 5)  # -8
        stack.add(DataType.FLOAT, 6)  # -4

        stack.move(-12)
        snapshop = stack.state()
        self.assertEqual(3, len(snapshop))
        self.assertEqual(3.0, snapshop.pop())
        self.assertEqual(2.0, snapshop.pop())
        self.assertEqual(1.0, snapshop.pop())

    def test_move_zero(self):
        stack = Stack()
        stack.add(DataType.FLOAT, 1)  # -24
        stack.add(DataType.FLOAT, 2)  # -20
        stack.add(DataType.FLOAT, 3)  # -16
        stack.add(DataType.FLOAT, 4)  # -12
        stack.add(DataType.FLOAT, 5)  # -8
        stack.add(DataType.FLOAT, 6)  # -4

        stack.move(0)
        snapshop = stack.state()
        self.assertEqual(6, len(snapshop))
        self.assertEqual(6.0, snapshop.pop())
        self.assertEqual(5.0, snapshop.pop())
        self.assertEqual(4.0, snapshop.pop())
        self.assertEqual(3.0, snapshop.pop())
        self.assertEqual(2.0, snapshop.pop())
        self.assertEqual(1.0, snapshop.pop())

    def test_copy_down_single(self):
        stack = Stack()
        stack.add(DataType.FLOAT, 1)  # -24
        stack.add(DataType.FLOAT, 2)  # -20
        stack.add(DataType.FLOAT, 3)  # -16
        stack.add(DataType.FLOAT, 4)  # -12
        stack.add(DataType.FLOAT, 5)  # -8
        stack.add(DataType.FLOAT, 6)  # -4

        stack.copy_down(-12, 4)

        self.assertEqual(6, stack.peek(-12))

    def test_copy_down_many(self):
        stack = Stack()
        stack.add(DataType.FLOAT, 1)  # -24
        stack.add(DataType.FLOAT, 2)  # -20
        stack.add(DataType.FLOAT, 3)  # -16
        stack.add(DataType.FLOAT, 4)  # -12
        stack.add(DataType.FLOAT, 5)  # -8
        stack.add(DataType.FLOAT, 6)  # -4

        stack.copy_down(-24, 12)
        print(stack.state())

        self.assertEqual(4, stack.peek(-24))
        self.assertEqual(5, stack.peek(-20))
        self.assertEqual(6, stack.peek(-16))



from __future__ import annotations

from typing import Optional

from pykotor.resource.formats.erf import ERF, ERFType
from pykotor.resource.type import ResourceType, TARGET_TYPES, ResourceReader, SOURCE_TYPES, ResourceWriter, autoclose


class ERFBinaryReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._erf: Optional[ERF] = None

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> ERF:
        self._erf = ERF()

        file_type = self._reader.read_string(4)
        file_version = self._reader.read_string(4)

        if not any(x for x in ERFType if x.value == file_type):
            raise ValueError("Not a valid ERF file.")

        if file_version != "V1.0":
            raise ValueError("The ERF version that was loaded is unsupported.")

        self._reader.skip(8)
        entry_count = self._reader.read_uint32()
        self._reader.skip(4)
        offset_to_keys = self._reader.read_uint32()
        offset_to_resources = self._reader.read_uint32()

        resrefs = []
        resids = []
        restypes = []
        self._reader.seek(offset_to_keys)
        for i in range(entry_count):
            resrefs.append(self._reader.read_string(16))
            resids.append(self._reader.read_uint32())
            restypes.append(self._reader.read_uint16())
            self._reader.skip(2)

        resoffsets = []
        ressizes = []
        self._reader.seek(offset_to_resources)
        for i in range(entry_count):
            resoffsets.append(self._reader.read_uint32())
            ressizes.append(self._reader.read_uint32())

        for i in range(entry_count):
            self._reader.seek(resoffsets[i])
            resdata = self._reader.read_bytes(ressizes[i])
            self._erf.set(resrefs[i], ResourceType.from_id(restypes[i]), resdata)

        return self._erf


class ERFBinaryWriter(ResourceWriter):
    FILE_HEADER_SIZE = 160
    KEY_ELEMENT_SIZE = 24
    RESOURCE_ELEMENT_SIZE = 8

    def __init__(
            self,
            erf: ERF,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self.erf = erf

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        entry_count = len(self.erf)
        offset_to_keys = ERFBinaryWriter.FILE_HEADER_SIZE
        offset_to_resources = offset_to_keys + ERFBinaryWriter.KEY_ELEMENT_SIZE * entry_count

        self._writer.write_string(self.erf.erf_type.value)
        self._writer.write_string("V1.0")
        self._writer.write_uint32(0)
        self._writer.write_uint32(0)
        self._writer.write_uint32(entry_count)
        self._writer.write_uint32(0)
        self._writer.write_uint32(offset_to_keys)
        self._writer.write_uint32(offset_to_resources)
        self._writer.write_uint32(0)
        self._writer.write_uint32(0)
        self._writer.write_uint32(0xFFFFFFFF)
        self._writer.write_bytes(b'\0' * 116)

        resid = 0
        for resource in self.erf:
            self._writer.write_string(resource.resref.get(), string_length=16)
            self._writer.write_uint32(resid)
            self._writer.write_uint16(resource.restype.type_id)
            self._writer.write_uint16(0)
            resid += 1

        data_offset = offset_to_resources + ERFBinaryWriter.RESOURCE_ELEMENT_SIZE * entry_count
        for resource in self.erf:
            self._writer.write_uint32(data_offset)
            self._writer.write_uint32(len(resource.data))
            data_offset += len(resource.data)

        for resource in self.erf:
            self._writer.write_bytes(resource.data)



"""
This module handles classes relating to editing TLK files.
"""
from __future__ import annotations

from typing import List

from pykotor.common.language import Language
from pykotor.common.misc import ResRef
from pykotor.resource.type import ResourceType


class TLK:
    BINARY_TYPE = ResourceType.TLK

    def __init__(
            self
    ):
        self.entries: List[TLKEntry] = []
        self.language: Language = Language.ENGLISH

    def __len__(
            self
    ):
        """
        Returns the number of stored entries.
        """
        return len(self.entries)

    def __iter__(
            self
    ):
        """
        Iterates through the stored entry with each iteration yielding a stringref and the corresponding entry data.
        """
        for stringref, entry in enumerate(self.entries):
            yield stringref, entry

    def __getitem__(
            self,
            item
    ):
        """
        Returns an entry for the specified stringref.

        Args:
            item: The stringref.

        Raises:
            IndexError: If the stringref does not exist.

        Returns:
            The corresponding TLKEntry.
        """
        if not isinstance(item, int):
            return NotImplemented
        return self.entries[item]

    def get(
            self,
            stringref: int
    ) -> TLKEntry:
        """
        Returns an entry for the specified stringref if it exists, otherwise returns None.

        Args:
            stringref: The stringref.

        Returns:
            The corresponding TLKEntry or None.
        """
        return self.entries[stringref] if 0 <= stringref < len(self) else None

    def add(
            self,
            text: str,
            sound_resref: str = ""
    ) -> int:
        entry = TLKEntry(text, ResRef(sound_resref))
        self.entries.append(entry)
        return len(self.entries) - 1

    def resize(
            self,
            size: int
    ) -> None:
        """
        Resizes the number of entries to the specified size.

        Args:
            size: The new number of entries.
        """
        if len(self) > size:
            self.entries = self.entries[:size]
        else:
            self.entries.extend([TLKEntry("", ResRef.from_blank()) for _ in range(len(self), size)])


class TLKEntry:
    def __init__(
            self,
            text: str,
            voiceover: ResRef
    ):
        self.text: str = text
        self.voiceover: ResRef = voiceover

    def __eq__(
            self,
            other
    ):
        """
        Returns True if the text and voiceover match.
        """
        if not isinstance(other, TLKEntry):
            return NotImplemented
        return other.text == self.text and other.voiceover == self.voiceover



from unittest import TestCase

from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.ifo import construct_ifo, dismantle_ifo

TEST_FILE = "../../files/test.ifo"


class TestIFO(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        ifo = construct_ifo(gff)
        self.validate_io(ifo)

        gff = dismantle_ifo(ifo)
        ifo = construct_ifo(gff)
        self.validate_io(ifo)

    def validate_io(self, ifo):
        self.assertEqual(b'R:\xe5\x9e\xe3sq\x1d\x0f\xf0i\x9c\xb9a\x9f\xa7', ifo.mod_id)
        self.assertEqual(2, ifo.creator_id)
        self.assertEqual(3, ifo.version)
        self.assertEqual("262", ifo.vo_id)
        self.assertEqual(0, ifo.expansion_id)
        self.assertEqual(83947, ifo.mod_name.stringref)
        self.assertEqual("262TEL", ifo.tag)
        self.assertEqual("", ifo.hak)
        self.assertEqual(-1, ifo.description.stringref)
        self.assertEqual("262tel", ifo.identifier)
        self.assertEqual(2.5811009407043457, ifo.entry_position.x)
        self.assertEqual(41.46979522705078, ifo.entry_position.y)
        self.assertEqual(21.372770309448242, ifo.entry_position.z)
        self.assertEqual(6, ifo.dawn_hour)
        self.assertEqual(18, ifo.dusk_hour)
        self.assertEqual(2, ifo.time_scale)
        self.assertEqual(6, ifo.start_month)
        self.assertEqual(1, ifo.start_day)
        self.assertEqual(13, ifo.start_hour)
        self.assertEqual(1372, ifo.start_year)
        self.assertEqual(10, ifo.xp_scale)
        self.assertEqual("heartbeat", ifo.on_heartbeat)
        self.assertEqual("load", ifo.on_load)
        self.assertEqual("start", ifo.on_start)
        self.assertEqual("enter", ifo.on_enter)
        self.assertEqual("leave", ifo.on_leave)
        self.assertEqual("activate", ifo.on_activate_item)
        self.assertEqual("acquire", ifo.on_acquire_item)
        self.assertEqual("user", ifo.on_user_defined)
        self.assertEqual("unacquire", ifo.on_unacquire_item)
        self.assertEqual("death", ifo.on_player_death)
        self.assertEqual("dying", ifo.on_player_dying)
        self.assertEqual("levelup", ifo.on_player_levelup)
        self.assertEqual("spawn", ifo.on_player_respawn)
        self.assertEqual("", ifo.on_player_rest)
        self.assertEqual("", ifo.start_movie)
        self.assertAlmostEqual(-89.99999, ifo.entry_direction, 4)



from unittest import TestCase

from pykotor.common.geometry import Polygon2, Vector2, Vector3, Vector4, Face


class TestVector2(TestCase):
    def test_unpacking(self):
        source = Vector2(1.2, 2.3)
        x, y = source
        self.assertEqual(x, 1.2)
        self.assertEqual(y, 2.3)

    def test_from_vector2(self):
        source = Vector2(1.2, 2.3)
        vec2 = Vector2.from_vector2(source)
        self.assertEqual(vec2.x, 1.2)
        self.assertEqual(vec2.y, 2.3)

    def test_from_vector3(self):
        source = Vector3(1.2, 2.3, 3.4)
        vec2 = Vector2.from_vector3(source)
        self.assertEqual(vec2.x, 1.2)
        self.assertEqual(vec2.y, 2.3)

    def test_from_vector4(self):
        source = Vector4(1.2, 2.3, 3.4, 5.6)
        vec2 = Vector2.from_vector4(source)
        self.assertEqual(vec2.x, 1.2)
        self.assertEqual(vec2.y, 2.3)


class TestVector3(TestCase):
    def test_unpacking(self):
        source = Vector3(1.2, 2.3, 3.4)
        x, y, z = source
        self.assertEqual(x, 1.2)
        self.assertEqual(y, 2.3)
        self.assertEqual(z, 3.4)

    def test_from_vector2(self):
        source = Vector2(1.2, 2.3)
        vec3 = Vector3.from_vector2(source)
        self.assertEqual(vec3.x, 1.2)
        self.assertEqual(vec3.y, 2.3)
        self.assertEqual(vec3.z, 0.0)

    def test_from_vector3(self):
        source = Vector3(1.2, 2.3, 3.4)
        vec3 = Vector3.from_vector3(source)
        self.assertEqual(vec3.x, 1.2)
        self.assertEqual(vec3.y, 2.3)
        self.assertEqual(vec3.z, 3.4)

    def test_from_vector4(self):
        source = Vector4(1.2, 2.3, 3.4, 5.6)
        vec3 = Vector3.from_vector4(source)
        self.assertEqual(vec3.x, 1.2)
        self.assertEqual(vec3.y, 2.3)
        self.assertEqual(vec3.z, 3.4)


class TestVector4(TestCase):
    def test_unpacking(self):
        source = Vector4(1.2, 2.3, 3.4, 4.5)
        x, y, z, w = source
        self.assertEqual(x, 1.2)
        self.assertEqual(y, 2.3)
        self.assertEqual(z, 3.4)
        self.assertEqual(w, 4.5)

    def test_from_vector2(self):
        source = Vector2(1.2, 2.3)
        vec4 = Vector4.from_vector2(source)
        self.assertEqual(vec4.x, 1.2)
        self.assertEqual(vec4.y, 2.3)
        self.assertEqual(vec4.z, 0.0)
        self.assertEqual(vec4.w, 0.0)

    def test_from_vector3(self):
        source = Vector3(1.2, 2.3, 3.4)
        vec4 = Vector4.from_vector3(source)
        self.assertEqual(vec4.x, 1.2)
        self.assertEqual(vec4.y, 2.3)
        self.assertEqual(vec4.z, 3.4)
        self.assertEqual(vec4.w, 0.0)

    def test_from_vector4(self):
        source = Vector4(1.2, 2.3, 3.4, 5.6)
        vec4 = Vector4.from_vector4(source)
        self.assertEqual(vec4.x, 1.2)
        self.assertEqual(vec4.y, 2.3)
        self.assertEqual(vec4.z, 3.4)
        self.assertEqual(vec4.w, 5.6)

    def test_from_euler(self):
        q1 = Vector4.from_euler(0, 0, 0)
        q2 = Vector4.from_euler(90, 0, 0)
        q3 = Vector4.from_euler(0, 90, 0)
        q4 = Vector4.from_euler(0, 0, 90)

        self.assertAlmostEqual(0.0, q1.x, 1)
        self.assertAlmostEqual(0.0, q1.y, 1)
        self.assertAlmostEqual(0.0, q1.z, 1)
        self.assertAlmostEqual(1.0, q1.w, 1)

        self.assertAlmostEqual(0.7, q2.x, 1)
        self.assertAlmostEqual(0.0, q2.y, 1)
        self.assertAlmostEqual(0.0, q2.z, 1)
        self.assertAlmostEqual(0.7, q2.w, 1)

        self.assertAlmostEqual(0.0, q3.x, 1)
        self.assertAlmostEqual(0.7, q3.y, 1)
        self.assertAlmostEqual(0.0, q3.z, 1)
        self.assertAlmostEqual(0.7, q3.w, 1)

        self.assertAlmostEqual(0.0, q4.x, 1)
        self.assertAlmostEqual(0.0, q4.y, 1)
        self.assertAlmostEqual(0.7, q4.z, 1)
        self.assertAlmostEqual(0.7, q4.w, 1)


class TestFace(TestCase):
    def test_determine_z(self):
        v1 = Vector3(0.0, 0.0, 0.0)
        v2 = Vector3(1.0, 0.0, 1.0)
        v3 = Vector3(0.0, 1.0, 1.0)
        face = Face(v1, v2, v3)
        # At the points
        self.assertEqual(0.0, face.determine_z(0.0, 0.0))  # v1
        self.assertEqual(1.0, face.determine_z(1.0, 0.0))  # v2
        self.assertEqual(1.0, face.determine_z(0.0, 1.0))  # v3
        # Middle of each edge
        self.assertEqual(0.5, face.determine_z(0.5, 0.0))  # v1, v2
        self.assertEqual(0.5, face.determine_z(0.0, 0.5))  # v1, v3
        self.assertEqual(1.0, face.determine_z(0.5, 0.5))  # v2, v3
        # Centre of the face
        self.assertEqual(0.66, face.determine_z(0.33, 0.33))


class TestPolygon2(TestCase):
    def test_inside(self):
        poly = Polygon2([Vector2(0.0, 0.0), Vector2(0, 6), Vector2(6, 6), Vector2(6, 0), Vector2(3, 3)])

        # Inside
        self.assertTrue(poly.inside(Vector2(2.0, 4.0)))
        self.assertTrue(poly.inside(Vector2(5.0, 6.0)))

        # On the edge
        self.assertTrue(poly.inside(Vector2(0.0, 3.0)))
        self.assertFalse(poly.inside(Vector2(0.0, 3.0), False))

        # Outside
        self.assertFalse(poly.inside(Vector2(3.0, 0.0)))
        self.assertFalse(poly.inside(Vector2(3.0, 7.0)))

    def test_area(self):
        poly = Polygon2([Vector2(0.0, 0.0), Vector2(0, 6), Vector2(6, 6), Vector2(6, 0), Vector2(3, 3)])

        self.assertEqual(27.0, poly.area())




import platform
from unittest import TestCase

from pykotor.resource.formats.vis import VISAsciiReader, VIS
from pykotor.resource.formats.vis.vis_auto import write_vis, read_vis
from pykotor.resource.type import ResourceType


ASCII_TEST_FILE = "../../files/test.vis"
DOES_NOT_EXIST_FILE = "./thisfiledoesnotexist"
CORRUPT_ASCII_TEST_FILE = "../../files/test_corrupted.vis"


class TestVIS(TestCase):
    def test_binary_io(self):
        vis = VISAsciiReader(ASCII_TEST_FILE).load()
        self.validate_io(vis)

        data = bytearray()
        write_vis(vis, data, ResourceType.VIS)
        vis = read_vis(data)
        self.validate_io(vis)

    def validate_io(self, vis: VIS):
        self.assertTrue(vis.get_visible("room_01", "room_02"))
        self.assertTrue(vis.get_visible("room_01", "room_03"))
        self.assertTrue(vis.get_visible("room_01", "room_04"))

        self.assertTrue(vis.get_visible("room_02", "room_01"))
        self.assertFalse(vis.get_visible("room_02", "room_03"))
        self.assertFalse(vis.get_visible("room_02", "room_04"))

        self.assertTrue(vis.get_visible("room_03", "room_01"))
        self.assertTrue(vis.get_visible("room_03", "room_04"))
        self.assertFalse(vis.get_visible("room_03", "room_02"))

        self.assertTrue(vis.get_visible("room_04", "room_01"))
        self.assertTrue(vis.get_visible("room_04", "room_03"))
        self.assertFalse(vis.get_visible("room_04", "room_02"))

    def test_read_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, read_vis, ".")
        else:
            self.assertRaises(IsADirectoryError, read_vis, ".")
        self.assertRaises(FileNotFoundError, read_vis, DOES_NOT_EXIST_FILE)
        self.assertRaises(ValueError, read_vis, CORRUPT_ASCII_TEST_FILE)

    def test_write_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, write_vis, VIS(), ".", ResourceType.VIS)
        else:
            self.assertRaises(IsADirectoryError, write_vis, VIS(), ".", ResourceType.VIS)
        self.assertRaises(ValueError, write_vis, VIS(), ".", ResourceType.INVALID)



from unittest import TestCase

from pykotor.common.misc import EquipmentSlot
from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.utc import construct_utc, dismantle_utc, UTC

TEST_FILE = "../../files/test.utc"


class TestUTC(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        utc = construct_utc(gff)
        self.validate_io(utc)

        gff = dismantle_utc(utc)
        utc = construct_utc(gff)
        self.validate_io(utc)

    def validate_io(self, utc: UTC):
        self.assertEqual(636, utc.appearance_id)
        self.assertEqual(1, utc.body_variation)
        self.assertEqual(120.0, utc.blindspot)
        self.assertEqual(10, utc.charisma)
        self.assertEqual(1.0, utc.challenge_rating)
        self.assertEqual("comment", utc.comment)
        self.assertEqual(10, utc.constitution)
        self.assertEqual("coorta", utc.conversation)
        self.assertEqual(1, utc.fp)
        self.assertEqual(8, utc.current_hp)
        self.assertEqual(10, utc.dexterity)
        self.assertTrue(utc.disarmable)
        self.assertEqual(5, utc.faction_id)
        self.assertEqual(76046, utc.first_name.stringref)
        self.assertEqual(1, utc.max_fp)
        self.assertEqual(2, utc.gender_id)
        self.assertEqual(50, utc.alignment)
        self.assertEqual(8, utc.hp)
        self.assertTrue(utc.hologram)
        self.assertTrue(utc.ignore_cre_path)
        self.assertEqual(10, utc.intelligence)
        self.assertTrue(utc.interruptable)
        self.assertTrue(utc.is_pc)
        self.assertEqual(123, utc.last_name.stringref)
        self.assertEqual(8, utc.max_hp)
        self.assertTrue(utc.min1_hp)
        self.assertEqual(3, utc.multiplier_set)
        self.assertEqual(1, utc.natural_ac)
        self.assertTrue(utc.no_perm_death)
        self.assertTrue(utc.not_reorienting)
        self.assertTrue(utc.party_interact)
        self.assertEqual(11, utc.perception_id)
        self.assertTrue(utc.plot)
        self.assertEqual(1, utc.portrait_id)
        self.assertEqual(6, utc.race_id)
        self.assertEqual("k_def_attacked01", utc.on_attacked)
        self.assertEqual("k_def_damage01", utc.on_damaged)
        self.assertEqual("k_def_death01", utc.on_death)
        self.assertEqual("k_def_dialogue01", utc.on_dialog)
        self.assertEqual("k_def_disturb01", utc.on_disturbed)
        self.assertEqual("k_def_endconv", utc.on_end_dialog)
        self.assertEqual("k_def_combend01", utc.on_end_round)
        self.assertEqual("k_def_heartbt01", utc.on_heartbeat)
        self.assertEqual("k_def_blocked01", utc.on_blocked)
        self.assertEqual("k_def_percept01", utc.on_notice)
        self.assertEqual("k_def_spawn01", utc.on_spawn)
        self.assertEqual("k_def_spellat01", utc.on_spell)
        self.assertEqual("k_def_userdef01", utc.on_user_defined)
        self.assertEqual(46, utc.soundset_id)
        self.assertEqual(10, utc.strength)
        self.assertEqual(1, utc.subrace_id)
        self.assertEqual("Coorta", utc.tag)
        self.assertEqual("n_minecoorta", utc.resref)
        self.assertEqual(1, utc.texture_variation)
        self.assertEqual(7, utc.walkrate_id)
        self.assertEqual(10, utc.wisdom)
        self.assertEqual(1, utc.fortitude_bonus)
        self.assertEqual(1, utc.reflex_bonus)
        self.assertEqual(1, utc.willpower_bonus)

        self.assertEqual(2, len(utc.classes))
        self.assertEqual(1, utc.classes[1].class_id)
        self.assertEqual(3, utc.classes[1].class_level)
        self.assertEqual(2, len(utc.classes[1].powers))
        self.assertEqual(9, utc.classes[1].powers[0])

        self.assertEqual(2, len(utc.equipment.items()))
        self.assertEqual("mineruniform", utc.equipment[EquipmentSlot.ARMOR].resref)
        self.assertTrue(utc.equipment[EquipmentSlot.ARMOR].droppable)
        self.assertEqual("g_i_crhide008", utc.equipment[EquipmentSlot.HIDE].resref)
        self.assertFalse(utc.equipment[EquipmentSlot.HIDE].droppable)

        self.assertEqual(2, len(utc.feats))
        self.assertEqual(94, utc.feats[1])

        self.assertEqual(4, len(utc.inventory))
        self.assertTrue(utc.inventory[0].droppable)
        self.assertFalse(utc.inventory[1].droppable)
        self.assertEqual("g_w_thermldet01", utc.inventory[1].resref)

        self.assertEqual(1, utc.computer_use)
        self.assertEqual(2, utc.demolitions)
        self.assertEqual(3, utc.stealth)
        self.assertEqual(4, utc.awareness)
        self.assertEqual(5, utc.persuade)
        self.assertEqual(6, utc.repair)
        self.assertEqual(7, utc.security)
        self.assertEqual(8, utc.treat_injury)








from typing import Optional

from pykotor.resource.formats.ltr.ltr_data import LTR
from pykotor.resource.type import SOURCE_TYPES, ResourceReader, TARGET_TYPES, ResourceWriter


class LTRBinaryReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._lip: Optional[LTR] = None

    def load(
            self,
            auto_close: bool = True
    ) -> LTR:
        self._ltr = LTR()

        file_type = self._reader.read_string(4)
        file_version = self._reader.read_string(4)

        if file_type != "LTR ":
            raise TypeError("The file type that was loaded is invalid.")

        if file_version != "V1.0":
            raise TypeError("The LTR version that was loaded is not supported.")

        letter_count = self._reader.read_uint8()
        if letter_count != 28:
            raise TypeError("LTR files that do not handle exactly 28 characters are not supported.")

        self._ltr._singles._start = [self._reader.read_single() for i in range(28)]
        self._ltr._singles._middle = [self._reader.read_single() for i in range(28)]
        self._ltr._singles._end = [self._reader.read_single() for i in range(28)]

        for i in range(28):
            self._ltr._doubles[i]._start = [self._reader.read_single() for j in range(28)]
            self._ltr._doubles[i]._middle = [self._reader.read_single() for j in range(28)]
            self._ltr._doubles[i]._end = [self._reader.read_single() for j in range(28)]

        for i in range(28):
            for j in range(28):
                self._ltr._triples[i][j]._start = [self._reader.read_single() for k in range(28)]
                self._ltr._triples[i][j]._middle = [self._reader.read_single() for k in range(28)]
                self._ltr._triples[i][j]._end = [self._reader.read_single() for k in range(28)]

        if auto_close:
            self._reader.close()

        return self._ltr


class LTRBinaryWriter(ResourceWriter):
    def __init__(
            self,
            ltr: LTR,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._ltr: LTR = ltr

    def write(
            self,
            auto_close: bool = True
    ) -> None:
        self._writer.write_string("LTR V1.0")
        self._writer.write_uint8(28)

        [self._writer.write_single(chance) for chance in self._ltr._singles._start]
        [self._writer.write_single(chance) for chance in self._ltr._singles._middle]
        [self._writer.write_single(chance) for chance in self._ltr._singles._end]

        for i in range(28):
            [self._writer.write_single(chance) for chance in self._ltr._doubles[i]._start]
            [self._writer.write_single(chance) for chance in self._ltr._doubles[i]._middle]
            [self._writer.write_single(chance) for chance in self._ltr._doubles[i]._end]

        for i in range(28):
            for j in range(28):
                [self._writer.write_single(chance) for chance in self._ltr._triples[i][j]._start]
                [self._writer.write_single(chance) for chance in self._ltr._triples[i][j]._middle]
                [self._writer.write_single(chance) for chance in self._ltr._triples[i][j]._end]

        if auto_close:
            self._writer.close()



import struct


def fix_audio(
        data: bytes
) -> bytes:
    b0x4 = struct.unpack('I', data[0:4])[0]
    b4x8 = struct.unpack('I', data[4:8])[0]
    b16x20 = struct.unpack('I', data[16:20])[0]
    if b0x4 == 1179011410 and b4x8 == 50 and b16x20 == 18:
        data = data[8:]
    elif b0x4 == 3294688255:
        data = data[470:]
    return data
